import itertools
import typing
from dataclasses import dataclass

import stk
import stko
from rdkit.Chem import AllChem as rdkit
from rdkit.Chem import rdMolTransforms

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble

from .generator import Generator


@dataclass
class TargetTorsion:
    smarts: str
    expected_num_atoms: int
    torsion_ids: tuple[int, int, int, int]


class TorsionScanner(Generator):
    """
    Generate conformers by scanning target torsions.

    """

    def __init__(
        self,
        target_torsions: TargetTorsion | tuple[TargetTorsion],
        angle_range: typing.Iterable[float],
    ):
        """
        Initialise generator.

        """

        if not isinstance(target_torsions, tuple):
            self._target_torsions = (target_torsions,)
        else:
            self._target_torsions = target_torsions

        self._angle_range = angle_range

    def generate_conformers(
        self,
        molecule: stk.BuildingBlock,
    ) -> Ensemble:
        # Optimise the initial ligand structure.
        molecule = stko.MMFF(ignore_inter_interactions=False).optimize(
            mol=molecule,
        )
        ensemble = Ensemble(base_molecule=molecule)
        rdkit_molecule = molecule.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_molecule)
        rdkit_properties = rdkit.MMFFGetMoleculeProperties(
            rdkit_molecule, mmffVariant="MMFF94s"
        )

        matched_torsions = {}
        atoms_to_be_constrained = set()
        for target in self._target_torsions:
            matches = rdkit_molecule.GetSubstructMatches(
                query=rdkit.MolFromSmarts(target.smarts),
            )

            for match in matches:
                assert len(match) == target.expected_num_atoms

                if not any(i in atoms_to_be_constrained for i in match):
                    for i in match:
                        atoms_to_be_constrained.add(i)

                    initial_torsion = rdMolTransforms.GetDihedralDeg(
                        rdkit_molecule.GetConformer(0),
                        match[target.torsion_ids[0]],
                        match[target.torsion_ids[1]],
                        match[target.torsion_ids[2]],
                        match[target.torsion_ids[3]],
                    )
                    key = tuple(match[i] for i in target.torsion_ids)
                    matched_torsions[key] = initial_torsion

        cid = 0
        test_molecule = molecule.clone()
        for target_angles in itertools.product(
            self._angle_range, repeat=len(matched_torsions)
        ):
            rdkit_molecule = test_molecule.to_rdkit_mol()
            rdkit.SanitizeMol(rdkit_molecule)
            rdkit_properties = rdkit.MMFFGetMoleculeProperties(rdkit_molecule)
            ff = rdkit.MMFFGetMoleculeForceField(
                rdkit_molecule,
                rdkit_properties,
            )
            for angle, torsion in zip(target_angles, matched_torsions):
                actual_angle = round(matched_torsions[torsion] + angle, 2)
                ff.MMFFAddTorsionConstraint(
                    torsion[0],
                    torsion[1],
                    torsion[2],
                    torsion[3],
                    False,
                    actual_angle - 0.1,
                    actual_angle + 0.1,
                    100.0,
                )
            ff.Minimize(maxIts=500)
            pos_mat = rdkit_molecule.GetConformer(-1).GetPositions()
            test_molecule = molecule.with_position_matrix(pos_mat)
            ensemble.add_conformer(
                conformer=Conformer(
                    molecule=test_molecule,
                    conformer_id=cid,
                    source="torsionscan",
                ),
            )
            cid += 1

        return ensemble

        # XTB CODE:
        # # ff = rdkit.MMFFGetMoleculeForceField(
        # # rdk_mol, rdk_props)
        # # ff.MMFFAddTorsionConstraint(
        # #     torsion_atom_ids[0],
        # #     torsion_atom_ids[1],
        # #     torsion_atom_ids[2],
        # #     torsion_atom_ids[3],
        # #     False,
        # #     angle - 0.2,
        # #     angle + 0.2,
        # #     1000.0,
        # # )
        # # ff.Minimize()

        # # new_conf = rdkit.Conformer(rdk_mol.GetNumAtoms())
        # # for i in range(rdk_mol.GetNumAtoms()):
        # #     new_conf.SetAtomPosition(
        # #         i, (
        # # rdk_mol.GetConformer(-1).GetAtomPosition(i))
        # #     )

        # # Set up directory.
        # xtb_dir = conformer_search_dir / f"{cname}_xtbopt"
        # out_file = xtb_dir / "output.txt"
        # if not os.path.exists(out_file):
        #     if os.path.exists(xtb_dir):
        #         shutil.rmtree(xtb_dir)
        #     os.mkdir(xtb_dir)
        #     init_dir = os.getcwd()
        #     os.chdir(xtb_dir)

        #     # Write constraint.inp file.
        #     # Xtb indexes start at 1. not 0.
        #     atoms1 = (
        #         f"{torsion_atom_ids_1[0]+1},"
        #         f"{torsion_atom_ids_1[1]+1},"
        #         f"{torsion_atom_ids_1[2]+1},"
        #         f"{torsion_atom_ids_1[3]+1}"
        #     )
        #     atoms2 = (
        #         f"{torsion_atom_ids_2[0]+1},"
        #         f"{torsion_atom_ids_2[1]+1},"
        #         f"{torsion_atom_ids_2[2]+1},"
        #         f"{torsion_atom_ids_2[3]+1}"
        #     )
        #     if overlapping:
        #         tcstr1 = f"dihedral: {atoms1},{float(angle1)}\n"
        #         tcstr2 = ""
        #     else:
        #         tcstr1 = f"dihedral: {atoms1},{float(angle1)}\n"
        #         tcstr2 = f"dihedral: {atoms2},{float(angle2)}\n"
        #     with open(xtb_dir / "input.inp", "w") as f:
        #         f.write(
        #             "$constrain\n"
        #             "force constant=1.0\n"
        #             f"{tcstr1}"
        #             f"{tcstr2}"
        #             "$end\n"
        #         )

        #     # Run xTB opt.
        #     conf_mol.write(xtb_dir / "input.xyz")
        #     xtb_path = (
        #         "/home/atarzia/miniconda3/envs/simple_het/bin/" "xtb"
        #     )
        #     cmd = (
        #         "ulimit -s unlimited ; "
        #         f"{xtb_path} "
        #         "input.xyz --gfnff "
        #         "--opt normal --parallel 4 "
        #         "--chrg 0 "
        #         "--input input.inp"
        #     )
        #     with open(out_file, "w") as f:
        #         # Note that sp.call will hold the program until
        #         # completion of the calculation.
        #         sp.call(
        #             cmd,
        #             stdin=sp.PIPE,
        #             stdout=f,
        #             stderr=sp.PIPE,
        #             # Shell is required for complex arguments.
        #             shell=True,
        #         )
        #     os.chdir(init_dir)

        # # Extract structure and energy.
        # # conf_mol = update_from_rdkit_conf(
        # #     stk_mol=opt_mol,
        # #     rdk_mol=rdk_mol,
        # #     conf_id=-1,
        # # )

        # try:
        #     conf_mol = conf_mol.with_structure_from_file(
        #         xtb_dir / "xtbopt.xyz"
        #     )
        #     with open(out_file, "r") as f:
        #         for line in f.readlines():
        #             if "TOTAL ENERGY" in line:
        #                 XtbEnergy = float(line.strip().split()[3])

        # except FileNotFoundError:
        #     logging.info(
        #         f"{xtb_dir} did not finish - ignoring this pair"
        #     )

        #     # This jumps back quite a bit in both angles, to
        #     # avoid just being at a clash point again.
        #     if overlapping:
        #         previous_cname = f"c_{angle1-20}_{int(angle2)}"
        #     else:
        #         previous_cname = f"c_{angle1-20}_{angle2-20}"

        #     previous_xtb_dir = (
        #         conformer_search_dir / f"{previous_cname}_xtbopt"
        #     )
        #     if not os.path.exists(previous_xtb_dir / "xtbopt.xyz"):
        #         if overlapping:
        #             previous_cname = f"c_{angle1-40}_{int(angle2)}"
        #         else:
        #             previous_cname = f"c_{angle1-40}_{angle2-40}"
        #         previous_xtb_dir = (
        #             conformer_search_dir / f"{previous_cname}_xtbopt"
        #         )

        #     if not os.path.exists(previous_xtb_dir / "xtbopt.xyz"):
        #         raise ValueError(
        #             "Had to go back many times, and none found."
        #             " This is broken!"
        #         )
        #     conf_mol = conf_mol.with_structure_from_file(
        #         previous_xtb_dir / "xtbopt.xyz"
        #     )
        #     continue



import typing

import numpy as np
import stk


class Selector:
    """
    Select atom ids in stk molecules by deleters.

    """

    def __init__(self):
        """
        Initialise Selector.

        """
        pass

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        raise NotImplementedError()

    def yield_stepwise(
        self,
        molecule: stk.BuildingBlock,
    ) -> typing.Iterator[tuple[int, ...]]:
        raise NotImplementedError()

    def get_atomic_positions(
        self,
        molecule: stk.BuildingBlock,
    ) -> tuple[np.ndarray, ...]:
        return tuple(
            molecule.get_atomic_positions(atom_ids=self.select_atoms(molecule))
        )


class NullSelector(Selector):
    """
    Selecter that does nothing.

    """

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple:
        return ()



from bbprep._internal.generators.etkdg import ETKDG
from bbprep._internal.generators.generator import Generator
from bbprep._internal.generators.torsion_scanner import (
    TargetTorsion,
    TorsionScanner,
)

__all__ = [
    "Generator",
    "ETKDG",
    "TorsionScanner",
    "TargetTorsion",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Distributed under the terms of the MIT License.

"""
Module for ensemble class.

Author: Andrew Tarzia

"""

from dataclasses import dataclass

import stk


@dataclass
class Conformer:
    molecule: stk.BuildingBlock
    conformer_id: int | None = None
    source: str | None = None


class Ensemble:
    def __init__(
        self,
        base_molecule: stk.Molecule,
    ):
        self._base_molecule = base_molecule
        self._molecule_num_atoms = base_molecule.get_num_atoms()
        self._conformers: dict[int, Conformer] = {}

    def get_num_conformers(self):
        return len(self._conformers)

    def add_conformer(self, conformer):
        if conformer.conformer_id is None:
            conf_id = self.get_num_conformers()
        else:
            conf_id = conformer.conformer_id

        assert conf_id not in self._conformers

        self._conformers[conf_id] = conformer

    def yield_conformers(self):
        for conf_id in self._conformers:
            yield self._conformers[conf_id]

    def get_base_molecule(self):
        return self._base_molecule

    def get_molecule_num_atoms(self):
        return self._molecule_num_atoms

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"num_confs={self.get_num_conformers()})"
        )

    def __repr__(self) -> str:
        return str(self)



import os
import pathlib

import stk
from bbprep import ClosestFGs


def test_closestfgs(molecule):
    """
    Test :class:`ClosestFGs`.

    Parameters:

        molecule:
            The molecule to modify.

    Returns:

        None : :class:`NoneType`

    """

    modified = ClosestFGs().modify(
        building_block=molecule.molecule,
        desired_functional_groups=molecule.desired_functional_groups,
    )
    modified_fgs = set(modified.get_functional_groups())
    original_fgs = tuple(molecule.molecule.get_functional_groups())
    fg_positions = []
    for idx, fg in enumerate(original_fgs):
        if fg in modified_fgs:
            assert idx in molecule.closest_ids
            fg_positions.append(
                molecule.molecule.get_centroid(atom_ids=fg.get_placer_ids())
            )

    xyz_string = stk.XyzWriter().to_string(molecule.molecule).split("\n")
    xyz_string[
        0
    ] = f"{int(xyz_string[0]) + molecule.desired_functional_groups}"
    for fgpos in fg_positions:
        xyz_string.append(
            f"He {round(fgpos[0], 2)} {round(fgpos[1], 2)} "
            f"{round(fgpos[2], 2)}"
        )

    path = pathlib.Path(os.path.dirname(os.path.realpath(__file__)))
    with open(path / f"dist_{molecule.name}_clo.xyz", "w") as f:
        f.write("\n".join(xyz_string))



import numpy as np
import stk

from .selector import Selector


class XCOMXSelector(Selector):
    """
    Get position of binder atoms with the molecule centroid in the middle.

    """

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        raise NotImplementedError(
            "This class can only provide positions, because one "
            "position is the centroid of the molecule."
        )

    def get_atomic_positions(
        self,
        molecule: stk.BuildingBlock,
    ) -> tuple[np.ndarray, ...]:
        assert molecule.get_num_functional_groups() == 2

        positions = []
        for fg in molecule.get_functional_groups():
            for id_ in fg.get_bonder_ids():  # type: ignore[attr-defined]
                positions.append(
                    next(molecule.get_atomic_positions(atom_ids=id_))
                )
        positions.insert(1, molecule.get_centroid())

        return tuple(positions)



import stk


class Modifier:
    """
    Modify a building block.

    """

    def __init__(self):
        """
        Initialise the process.

        """
        pass

    def modify(self, building_block: stk.BuildingBlock) -> stk.BuildingBlock:
        raise NotImplementedError()



import typing

import stk

from .selector import Selector


class BindersSelector(Selector):
    """
    Select atom ids in stk molecules by binders.

    """

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        assert molecule.get_num_functional_groups() > 0

        atoms = []
        for fg in molecule.get_functional_groups():
            for id_ in fg.get_bonder_ids():  # type: ignore[attr-defined]
                atoms.append(id_)

        return tuple(atoms)

    def yield_stepwise(
        self,
        molecule: stk.BuildingBlock,
    ) -> typing.Iterator[tuple[int, ...]]:
        assert molecule.get_num_functional_groups() > 0

        for fg in molecule.get_functional_groups():
            atoms = []
            for id_ in fg.get_bonder_ids():  # type: ignore[attr-defined]
                atoms.append(id_)
            yield tuple(atoms)



import stk

from bbprep._internal.ensemble.ensemble import Ensemble


class Generator:
    """
    Generate an ensemble of molecules.

    """

    def generate_conformers(
        self,
        molecule: stk.BuildingBlock,
    ) -> Ensemble:
        raise NotImplementedError()
